$alias fdopen='fdopen'(%val, %ref)
      include 'phigs.f1.h'		! get the HP-PHIGS aliases
      program CameraModel		! file "ZoomView.f"
      include 'phigs.f2.h'		! get the HP-PHIGS constants

      integer	WorkstnID		! workstation ID
      parameter	(WorkstnID=1)		! user-selectable value
      integer	ConnID			! communication channel ID
      integer	Error			! error indicator
      !--- define structure IDs ----------------------------------------------
      integer	SceneID, PlaneID, CircleID, CylinderID, ConeID
      integer	SphereID, SquareID, CubeID
      common	/StructureIDs/ SceneID, PlaneID, CircleID, CylinderID
      common	/StructureIDs/ ConeID, SphereID, SquareID, CubeID
      data	SceneID, PlaneID, CircleID, CylinderID /1, 2, 3, 4/
      data	ConeID, SphereID, SquareID, CubeID /5, 6, 7, 8/

      real	CenterX, CenterY	! center of image rotation
      data	CenterX /5./, CenterY /5./
      real	M1(4,4), M2(4,4), M3(4,4), M4(4,4), M(4,4)	! matrices
      real	X, Y, Z				! temporary position variables
      real	X1, Y1, Z1, X2, Y2, Z2		! position last time/this time
      real	VelX1, VelY1, VelZ1		! velocity last time
      real	VelX2, VelY2, VelZ2		! velocity this time
      real	AccelX1, AccelY1, AccelZ1	! acceleration last time
      real	AccelX2, AccelY2, AccelZ2	! acceleration this time
      integer	I			! loop control variable
      integer	fdopen			! to get file descriptor

      real	rad, deg			! type the statement function
      rad(deg)=	((deg)*3.14159265358979/180.)	! convert degrees to radians
      
      !=== open PHIGS and workstation ========================================
      call popph(fdopen(fnum(7), 'w'//char(0)), 0)	! open phigs
      call pue004('/dev/screen/phigs_window', ConnID)	! get connection ID
      call popwk(WorkstnID, ConnID, POIDDX)	! open workstation
      !=== define all needed structures ======================================
      !--- define the "floor" plane ------------------------------------------
      call popst(PlaneID)			! open structure 
      call Plane(0., 10., 1., 0., 10., 2.)	! define plane
      call pclst				! close structure
      !--- define the 2D shapes with supporting roles ------------------------
      call popst(CircleID)		! open structure
      call SupinePolygon(24)		! define a "circle"
      call pclst			! close structure
      call popst(SquareID)		! open structure
      call SupinePolygon(4)		! define a square
      call pclst			! close structure
      !--- define the cylinder -----------------------------------------------
      call popst(CylinderID)		! open structure 
      call Cylinder(1., .2, 1., .1)	! define cylinder
      call pclst			! close structure
      !--- define the cone ---------------------------------------------------
      call popst(ConeID)		! open structure
      call Cone(1., 2., .05)		! define cone
      call pclst			! close structure
      !--- define the sphere -------------------------------------------------
      call popst(SphereID)		! open structure 
      call Sphere(1., 5.)		! define sphere
      call pclst			! close structure
      !--- define the cube ---------------------------------------------------
      call popst(CubeID)		! open structure
      call Cube(sqrt(2.)/2, .1, 1., .1)	! define cube
      call pclst			! close structure
      !=== define the scene ==================================================
      call popst(SceneID)
      call psvwi(1)
      call pexst(PlaneID)
      !--- cone --------------------------------------------------------------
      call ptr3(5., 1., 0., Error, M)
      call pslmt3(M, PCREPL)
      call pexst(ConeID)
      !--- cylinders ---------------------------------------------------------
      call ptr3(2., 8., 0., Error, M)
      call pslmt3(M, PCREPL)
      call pexst(CylinderID)		! short, fat cylinder
      call ptr3(0., 0., -.5, Error, M)
      call psc3(.4, .4, 2., Error, M1)
      call AppendTransformation(M, M1)
      call prox(rad(90), Error, M1)
      call AppendTransformation(M, M1)
      call proz(rad(30), Error, M1)
      call AppendTransformation(M, M1)
      call ptr3(2.3, 8.3, 1.4, Error, M1)
      call AppendTransformation(M, M1)
      call pslmt3(M, PCREPL)
      call pexst(CylinderID)		! tall, thin cylinder
      !--- spheroids ---------------------------------------------------------
      call pbltm3(0.,0.,.0, 1.,1.,1., rad(45),0.,0., 1.,1.,1., Error, M)
      call pslmt3(M, PCREPL)
      call pexst(SphereID)		! sphere
      call pbltm3(0.,0.,.0, 8.,6.,2., 0.,0.,0., .5,.5,2., Error, M)
      call pslmt3(M, PCREPL)
      call pexst(SphereID)		! prolate spheroid
      call pbltm3(0.,0.,.0, 5.,1.,2.5, 0.,0.,0., 1.,1.,.5,
     +  Error, M)
      call pslmt3(M, PCREPL)
      call pexst(SphereID)		! oblate spheroid
      !--- cubes -------------------------------------------------------------
      call proz(rad(45), Error, M)
      call psc3(2., 1., 1., Error, M1)
      call AppendTransformation(M, M1)
      call prox(rad(90), Error, M1)
      call AppendTransformation(M, M1)
      call proz(rad(30), Error, M1)
      call AppendTransformation(M, M1)
      call ptr3(8., 3., .5, Error, M1)
      call AppendTransformation(M, M1)
      call pslmt3(M, PCREPL)
      call pexst(CubeID)		! bottom block
      call proz(rad(45), Error, M1)
      call proy(rad(90), Error, M2)
      call pcom3(M2, M1, Error, M3)
      call proz(rad(50), Error, M4)
      call pcom3(M4, M3, Error, M1)
      call ptr3(7.8, 2.5, 1.5, Error, M2)
      call pcom3(M2, M1, Error, M)
      call pslmt3(M, PCREPL)
      call pexst(CubeID)		! top cube
      call pclst			! close structure
      !=== move around/in the scene ==========================================
      call ppost(WorkstnID, SceneID, 1.0)	! post structure
      !--- go around the scene once ------------------------------------------
      do I=0,360,2
	  call PolarToRectangular(25., rad(I-90), rad(80), X, Y, Z)
	  call DefineCameraView(WorkstnID, 1,	! workstation id, view number
     +	    X+CenterX, Y+CenterY, Z,		! PRP (point looked from)
     +	    5., 5., 0.,				! VRP (point looked at)
     +	    30., PPERS,				! field of view, proj. type
     +      0., 0., 1.)				! up vector
	  call puwk(1, PPERFO)			! update workstation
      end do
      !--- look down on the scene from above ---------------------------------
      call psdus(WorkstnID, PASAP, PNIVE)	! set display update state...
      do I=80,4,-2				!  ...so we can avoid PUWKing.
	  call PolarToRectangular(25., rad(-90), rad(I), X, Y, Z)
	  call DefineCameraView(WorkstnID, 1,	! workstation id, view number
     +	    X+CenterX, Y+CenterY, Z,		! PRP (point looked from)
     +	    5., 5., 0.,				! VRP (point looked at)
     +	    30., PPERS,				! field of view, proj. type
     +      0., 0., 1.)				! up vector
      end do
      do I=4,90,2
	  call PolarToRectangular(25., rad(-90), rad(I), X, Y, Z)
	  call DefineCameraView(WorkstnID, 1,	! workstation id, view number
     +	    X+CenterX, Y+CenterY, Z,		! PRP (point looked from)
     +	    5., 5., 0.,				! VRP (point looked at)
     +	    30., PPERS,				! field of view, proj. type
     +      0., 0., 1.)				! up vector
      end do
      !--- move into the scene -----------------------------------------------
      X1=X+CenterX				! \  The point looked from at
      Y1=Y+CenterY				!  > the beginning of the  
      Z1=Z					! /  camera movement.
      X2=7.					! \  The point looked from at
      Y2=10.					!  > the end of the camera 
      Z2=1.					! /  movement.             
      do I=0,100
	  X=(100-I)*X1*.01+I*X2*.01		! \  Move linearly from point
	  Y=(100-I)*Y1*.01+I*Y2*.01		!  > 1 to point 2, in 101
	  Z=(100-I)*Z1*.01+I*Z2*.01		! /  easy steps
	  call DefineCameraView(WorkstnID, 1,	! workstation id, view number
     +	    X, Y, Z,				! PRP (point looked from)
     +	    5., 5., 0.,				! VRP (point looked at)
     +	    30., PPERS,				! field of view, proj. type
     +      0., 0., 1.)				! up vector
      end do
      !--- look up at the the cone/spheroid juncture -------------------------
      X1=5.					! \  The point looked at at
      Y1=5.					!  > the beginning of the  
      Z1=0.					! /  camera pan/tilt.      
      X2=5.					! \  The point looked at at
      Y2=1.					!  > the end of the camera 
      Z2=2.					! /  pan/tilt.             
      do I=0,50
	  X=(50-I)*X1*.02+I*X2*.02		! \  Move linearly from point
	  Y=(50-I)*Y1*.02+I*Y2*.02		!  > 1 to point 2, in 51
	  Z=(50-I)*Z1*.02+I*Z2*.02		! /  easy steps
	  call DefineCameraView(WorkstnID, 1,	! workstation id, view number
     +	    7., 10., 1.,			! PRP (point looked from)
     +	    X, Y, Z,				! VRP (point looked at)
     +	    30., PPERS,				! field of view, proj. type
     +      0., 0., 1.)				! up vector
      end do
      !--- zoom in on the cone/spheroid juncture -----------------------------
      do I=60,4,-1
	  call DefineCameraView(WorkstnID, 1,	! workstation id, view number
     +	    7., 10., 1.,			! PRP (point looked from)
     +	    X2, Y2, Z2,				! VRP (point looked at)
     +	    I*.5, PPERS,			! field of view, proj. type
     +      0., 0., 1.)				! up vector
      end do
      do I=4,270,2
	  call DefineCameraView(WorkstnID, 1,	! workstation id, view number
     +	    7., 10., 1.,			! PRP (point looked from)
     +	    X2, Y2, Z2,				! VRP (point looked at)
     +	    I*.5, PPERS,			! field of view, proj. type
     +      0., 0., 1.)				! up vector
      end do
      do I=270,120,-1
	  call DefineCameraView(WorkstnID, 1,	! workstation id, view number
     +	    7., 10., 1.,			! PRP (point looked from)
     +	    X2, Y2, Z2,				! VRP (point looked at)
     +	    I*.5, PPERS,			! field of view, proj. type
     +      0., 0., 1.)				! up vector
      end do
      !--- fly away as in an airplane ----------------------------------------
      AccelX1=0.
      AccelY1=0.
      AccelZ1=1.
      do I=0,1082
	  !--- devise a complex, but smooth, steadily rising flight path ---
	  X2=5-1.5*sin(rad(I*1.7))-.9*cos(rad(I*.34))
	  Y2=4+(I*.004+1)*cos(rad(I*1.3+15))+.5*sin(rad(I*1.6))
	  Z2=I*.003					
	  VelX2=X2-X1
	  VelY2=Y2-Y1
	  VelZ2=Z2-Z1
	  AccelX2=VelX2-VelX1
	  AccelY2=VelY2-VelY1
	  AccelZ2=VelZ2-VelZ1
	  if (I .gt. 3) then			! ignore anomalous values
	      call DefineCameraView(WorkstnID,1,! workstation id, view number
     +		X1, Y1, Z1,			! PRP (point looked from)
     +		X2, Y2, Z2,			! VRP (point looked at)
     +		90., PPERS,			! field of view, proj. type
     +		AccelX1, AccelY1, .003)		! up vector
	  end if
	  X1=X2
	  Y1=Y2
	  Z1=Z2
	  VelX1=VelX2
	  VelY1=VelY2
	  VelZ1=VelZ2
	  AccelX1=AccelX2
	  AccelY1=AccelY2
	  AccelZ1=AccelZ2
      end do
      call pclwk(WorkstnID)		! close workstation
      call pclph			! close PHIGS
      stop
      end
!*****************************************************************************
      subroutine Plane(Xmin, Xmax, dX, Ymin, Ymax, dY)
      real	Xmin, Xmax, Ymin, Ymax	! defines size of floor
      real	dX, dY			! size of tiles in floor
      include 'phigs.f2.h'		! get the HP-PHIGS constants

      real	X, Y			! loop control variables

      X=Xmin
      do while (X .le. Xmax+.00001)
	  call OneLine(X, Ymin, 0., X, Ymax, 0.)
	  X=X+dX
      end do
      Y=Ymin
      do while (Y .le. Ymax+.00001)
	  call OneLine(Xmin, Y, 0., Xmax, Y, 0.)
	  Y=Y+dY
      end do
      return
      end
!*****************************************************************************
      subroutine OneLine(X1, Y1, Z1, X2, Y2, Z2)
      real	X1, Y1, Z1, X2, Y2, Z2
      
      real	X(2), Y(2), Z(2)
      
      X(1)=X1
      Y(1)=Y1
      Z(1)=Z1
      X(2)=X2
      Y(2)=Y2
      Z(2)=Z2
      call ppl3(2, X, Y, Z)
      return
      end
!*****************************************************************************
      subroutine SupinePolygon(Sides)
      integer	Sides			! number of sides of polygon
      
      real	X(360), Y(360), Z(360)	! XYZ data for POLYLINE 3
      integer	dTheta, I, Index	! loop control variables
      real	Theta			! temporary variable

      real	rad, deg			! type the statement function
      rad(deg)=	((deg)*3.14159265358979/180.)	! convert degrees to radians

      dTheta=360/Sides			! degrees per side
      Index=1				! for array subscripts
      do I=0, 360, dTheta 		! for each side...
	  Theta=rad(I)			!   convert degrees to radians
	  X(Index)=cos(Theta)		!   \
	  Y(Index)=sin(Theta)		!    > calculate XYZ data
	  Z(Index)=0.			!   /
	  Index=Index+1			!   increment array subscript
      end do
      call ppl3(Sides+1, X, Y, Z)	! polyline 3
      return
      end
!*****************************************************************************
      subroutine Cylinder(Radius, dRadius, Height, dHeight)
      real	Radius			! radius of cylinder
      real	dRadius			! distance between circles in top/btm
      real	Height			! total height of cylinder
      real	dHeight			! distance between layers in cylinder
      include 'phigs.f2.h'		! get the HP-PHIGS constants

      integer	SceneID, PlaneID, CircleID, CylinderID, ConeID
      integer	SphereID, SquareID, CubeID
      common	/StructureIDs/ SceneID, PlaneID, CircleID, CylinderID
      common	/StructureIDs/ ConeID, SphereID, SquareID, CubeID

      real	M1(4,4), M2(4,4), M(4,4)! transformation matrices
      integer	Error			! error return variable
      real	R, Z			! loop control variables
      
      R=dRadius				! initial radius
      do while (R .le. Radius)
	  !--- do bottom first -----------------------------------------------
	  call psc3(R, R, 1., Error, M1)! scale 3
	  call pslmt3(M1, PCREPL)	! set local transformation 3
	  call pexst(CircleID)		! execute structure
	  !--- then do top ---------------------------------------------------
	  call ptr3(0., 0., Height, Error, M2)		! translate 3
	  call pcom3(M1, M2, Error, M)	! compose matrix 3
	  call pslmt3(M, PCREPL)	! set local transformation 3
	  call pexst(CircleID)		! execute structure
	  R=R+dRadius			! increment the radius
      end do
      Z=0.
      call psc3(Radius, Radius, 1., Error, M1)	! scale 3
      do while (Z .le. Height)
	  call ptr3(0., 0., Z, Error, M2)		! translate 3
	  call pcom3(M1, M2, Error, M)	! compose matrix 3
	  call pslmt3(M, PCREPL)	! set local transformation 3
	  call pexst(CircleID)		! execute structure
	  Z=Z+dHeight			! increment the height
      end do
      return
      end
!*****************************************************************************
      subroutine Cone(Radius, Height, dHeight)
      real	Radius			! radius of sphere
      real	Height			! height of whole cone
      real	dHeight			! distance between rings
      include 'phigs.f2.h'		! get the HP-PHIGS constants

      integer	SceneID, PlaneID, CircleID, CylinderID, ConeID
      integer	SphereID, SquareID, CubeID
      common	/StructureIDs/ SceneID, PlaneID, CircleID, CylinderID
      common	/StructureIDs/ ConeID, SphereID, SquareID, CubeID

      real	M1(4,4), M2(4,4), M(4,4)! transformation matrices
      integer	Error			! error return variable
      real	Z, R			! loop control variable

      Z=dHeight				! initial height
      do while (Z .le. Height)
	  R=Radius*((Height-Z)/Height)
	  call psc3(R, R, 1., Error, M1)! scale 3
	  call ptr3(0., 0., Z, Error, M2)	! translate 3
	  call pcom3(M1, M2, Error, M)	! compose matrix 3
	  call pslmt3(M, PCREPL)	! set local transformation 3
	  call pexst(CircleID)		! execute structure
	  Z=Z+dHeight			! increment the height
      end do
      return
      end
!*****************************************************************************
      subroutine Sphere(Radius, dPhi)
      real	Radius			! radius of sphere
      real	dPhi			! angular distance between latitudes
      include 'phigs.f2.h'		! get the HP-PHIGS constants

      integer	SceneID, PlaneID, CircleID, CylinderID, ConeID
      integer	SphereID, SquareID, CubeID
      common	/StructureIDs/ SceneID, PlaneID, CircleID, CylinderID
      common	/StructureIDs/ ConeID, SphereID, SquareID, CubeID

      real	M1(4,4), M2(4,4), M(4,4)! transformation matrices
      integer	Error			! error return variable
      real	Phi, R, Z		! loop control variables

      real	rad, deg			! type the statement function
      rad(deg)=	((deg)*3.14159265358979/180.)	! convert degrees to radians
      
      Phi=dPhi				! initial radius
      do while (Phi .lt. 180.)
	  R=Radius*sin(rad(Phi))
	  call psc3(R, R, 1., Error, M1)! scale 3
	  Z=Radius*cos(rad(Phi))
	  call ptr3(0., 0., Z, Error, M2)	! translate 3
	  call pcom3(M1, M2, Error, M)	! compose matrix 3
	  call pslmt3(M, PCREPL)	! set local transformation 3
	  call pexst(CircleID)		! execute structure
	  Phi=Phi+dPhi			! increment Phi
      end do
      return
      end
!*****************************************************************************
      subroutine Cube(Distance, dDistance, Height, dHeight)
      real	Distance		! center-to-corner distance
      real	dDistance		! distance between vertices in top/btm
      real	Height			! total height of cube
      real	dHeight			! distance between layers in cube
      include 'phigs.f2.h'		! get the HP-PHIGS constants

      integer	SceneID, PlaneID, CircleID, CylinderID, ConeID
      integer	SphereID, SquareID, CubeID
      common	/StructureIDs/ SceneID, PlaneID, CircleID, CylinderID
      common	/StructureIDs/ ConeID, SphereID, SquareID, CubeID

      real	M1(4,4), M2(4,4), M(4,4)! transformation matrices
      integer	Error			! error return variable
      real	Dist, Z			! loop control variables
      
      Dist=dDistance				! initial radius
      do while (Dist .le. Distance)
	  !--- do bottom first -----------------------------------------------
	  call psc3(Dist, Dist, 1., Error, M1)! scale 3
	  call pslmt3(M1, PCREPL)	! set local transformation 3
	  call pexst(SquareID)		! execute structure
	  !--- then do top ---------------------------------------------------
	  call ptr3(0., 0., Height, Error, M2)		! translate 3
	  call pcom3(M1, M2, Error, M)	! compose matrix 3
	  call pslmt3(M, PCREPL)	! set local transformation 3
	  call pexst(SquareID)		! execute structure
	  Dist=Dist+dDistance		! increment the "radius"
      end do
      Z=0.
      call psc3(Distance, Distance, 1., Error, M1)	! scale 3
      do while (Z .le. Height)
	  call ptr3(0., 0., Z, Error, M2)		! translate 3
	  call pcom3(M1, M2, Error, M)	! compose matrix 3
	  call pslmt3(M, PCREPL)	! set local transformation 3
	  call pexst(SquareID)		! execute structure
	  Z=Z+dHeight			! increment the height
      end do
      return
      end
!*****************************************************************************
      subroutine DefineCameraView(WorkstnID, ViewNo, PRPx, PRPy, PRPz,
     +  VRPx, VRPy, VRPz, FieldOfView, ProjType, VUPx, VUPy, VUPz)
      integer	WorkstnID, ViewNo	! workstation id, view to be defined
      real	PRPx, PRPy, PRPz	! Proj. Ref. Point (point looked from)
      real	VRPx, VRPy, VRPz	! View Ref. Point (point looked at)
      real	FieldOfView		! "zoomness" in degrees
      integer	ProjType		! parallel or perspective?
      real	VUPx, VUPy, VUPz	! view up vector
      include 'phigs.f2.h'		! get the HP-PHIGS constants

      real	VPNx, VPNy, VPNz	! view plane normal
      real	FPD, BPD		! front/back place distance
      real	Mapping(4,4)		! view mapping matrix
      real	Orientation(4,4)	! view orientation matrix
      real	Distance		! distance from PRP to VRP
      real	Wndw, Window(4)		! window limits
      real	Viewport(6)		! viewport limits
      real	ClipLimits(6)		! clip limits
      data	Viewport   /0.0, 1.0, 0.0, 1.0, 0.0, 1.0/
      data	ClipLimits /0.0, 1.0, 0.0, 1.0, 0.0, 1.0/
      integer	Error			! error return variable
      
      real	rad, deg			! type the statement function
      rad(deg)=	((deg)*3.14159265358979/180.)	! convert degrees to radians

      VPNx=PRPx-VRPx			! \  View Plane Normal determined by
      VPNy=PRPy-VRPy			!  > Projection Reference Point and
      VPNz=PRPz-VRPz			! /  View Reference Point.
      call pevom3(VRPx, VRPy, VRPz, VPNx, VPNy, VPNz, VUPx, VUPy, VUPz,
     +  Error, Orientation)
      if (Error .ne. 0) then
	  print *, "Error", Error, " in pevom3; terminating."
	  stop
      end if
      Distance=sqrt((PRPx-VRPx)**2+(PRPy-VRPy)**2+(PRPz-VRPz)**2)
      Wndw=Distance*tan(rad(FieldOfView)/2)	! FOV determines window size
      Window(1)=-Wndw			! \
      Window(2)=Wndw			!  \ This assumes a square window
      Window(3)=-Wndw			!  / (aspect ratio=1.00).
      Window(4)=Wndw			! /
      FPD=Distance-0.01			! right in front of eye point
      BPD=-450*Distance			! virtually infinite
      call pevmm3(Window, Viewport, ProjType, 0., 0., Distance,
     +  0., BPD, FPD, Error, Mapping)
      if (Error .ne. 0) then
	  print *, "Error", Error, " in pevmm3; terminating."
	  stop
      end if
      call psvwr3(WorkstnID, ViewNo, Orientation, Mapping, ClipLimits,
     +  PCLIP, PCLIP, PCLIP)
      return
      end
!*****************************************************************************
      subroutine PolarToRectangular(R, Theta, Phi, X, Y, Z)
      real	R, Theta, Phi		! input: 3D polar (spherical) coords
      real	X, Y, Z			! output: 3D rect. (Cartesian) coords
      
      X=R*sin(Phi)*cos(Theta)
      Y=R*sin(Phi)*sin(Theta)
      Z=R*cos(Phi)
      return
      end
!*****************************************************************************
      subroutine AppendTransformation(MainTransform, Appendix)
      real	MainTransform(4,4)	! the transformation to be appended to
      real	Appendix(4,4)		! the transformation to be appended
      
      real	Temp(4,4)		! temporary matrix holder
      integer	Error			! error indicator
      integer	I, J			! loop control variables
      
      call pcom3(Appendix, MainTransform, Error, Temp)
      do I=1,4					! \
	  do J=1,4				!  \  Copy the result matrix
	      MainTransform(I,J)=Temp(I,J)	!   > back into the first
	  end do				!  /  argument matrix.
      end do					! /
      return
      end
